package com.agentfactory.agentspeak.interpreter;

import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Observable;

import com.agentfactory.agentspeak.interpreter.control.BasicCommitmentSelector;
import com.agentfactory.agentspeak.interpreter.control.BasicEventSelector;
import com.agentfactory.agentspeak.interpreter.control.BasicOptionGenerator;
import com.agentfactory.agentspeak.interpreter.control.BasicOptionSelector;
import com.agentfactory.agentspeak.interpreter.control.ICommitmentSelector;
import com.agentfactory.agentspeak.interpreter.control.IEventSelector;
import com.agentfactory.agentspeak.interpreter.control.IOptionGenerator;
import com.agentfactory.agentspeak.interpreter.control.IOptionSelector;
import com.agentfactory.agentspeak.interpreter.mentalstate.ASEvent;
import com.agentfactory.agentspeak.interpreter.mentalstate.BeliefBase;
import com.agentfactory.agentspeak.interpreter.mentalstate.Commitment;
import com.agentfactory.agentspeak.interpreter.mentalstate.IPlanStep;
import com.agentfactory.agentspeak.interpreter.mentalstate.PlanRule;
import com.agentfactory.logic.update.interpreter.AbstractLanguageAgent;
import com.agentfactory.logic.update.interpreter.Action;
import com.agentfactory.logic.update.interpreter.CoreUtilities;
import com.agentfactory.logic.update.interpreter.Sensor;
import com.agentfactory.logic.update.lang.IFormula;
import com.agentfactory.logic.update.lang.Predicate;
import com.agentfactory.logic.update.reasoner.IQueryEngine;
import com.agentfactory.logic.update.reasoner.QueryResult;
import com.agentfactory.logic.update.reasoner.ResolutionBasedQueryEngine;
import com.agentfactory.logic.update.util.LogicException;

public class AbstractAgentSpeakAgent extends AbstractLanguageAgent {
	/*********************************************************************
	 * Core Interpreter Data Structures
	 *********************************************************************/
	protected List<ASEvent> eventQueue;
	protected BeliefBase beliefs;
	protected IQueryEngine queryEngine;
	protected List<Commitment> commitments;
	protected Map<String, List<PlanRule>> ruleEventMap;

	/*********************************************************************
	 * Current State of Algorithm (i.e. what was selected in the loop
	 *********************************************************************/
	private ASEvent event;
	private List<PlanRule> options;
	private PlanRule option;
	
	/*********************************************************************
	 * Interpreter Selection Algorithms
	 *********************************************************************/
	private IEventSelector eventSelector;
	private IOptionGenerator optionGenerator;
	private IOptionSelector optionSelector;
	private ICommitmentSelector commitmentSelector;
	
	public AbstractAgentSpeakAgent(String name) {
		super(name);

		// Create an event queue and add an "initialized" event
		// that can be used to hook initial behaviours that should
		// be performed by the agent.
		eventQueue = new LinkedList<ASEvent>();
		eventQueue.add(new ASEvent(ASEvent.ADDITION, (Predicate) Utilities.factory.createFormula("initialized")));
		
		beliefs = new BeliefBase();
		queryEngine = new ResolutionBasedQueryEngine();
		queryEngine.addSource(beliefs);
		
		ruleEventMap = new HashMap<String, List<PlanRule>>();
		
		commitments = new LinkedList<Commitment>();
		
		eventSelector = new BasicEventSelector(eventQueue);
		optionGenerator = new BasicOptionGenerator(ruleEventMap);
		optionSelector = new BasicOptionSelector(queryEngine);
		commitmentSelector = new BasicCommitmentSelector();
		
		installBasicAPI();		
	}
	
	@Override
	public void execute() {
		this.retrieveNewMessages();
		beliefs.wipe("percept");
		this.senseEnvironment();
		this.getOutbox().clear();
		
		// Part B: Handle New Events (if one exists)
		event = eventSelector.next();
		if (event != null) {
			// 1. Generate a set of options
			options = optionGenerator.generate(event);
				
			// 2. Select an option & commit to it
			option = optionSelector.select(options);
			if (option != null) {
				if (event.getSource() != null) {
					Commitment commitment = (Commitment) event.getSource();
					commitment.addPlan(option);
					commitment.resume();
				} else {	
					commitments.add(new Commitment(event, option));
				}
			}
		}
		
		// Part C: Handle Existing Commitments (if any exist)
		if (!commitments.isEmpty()) {
			Commitment commitment = commitmentSelector.select(commitments);
			IPlanStep planStep = commitment.getNextActivity();
			if (planStep == null) {
				// No Action => Commitment is completed
				commitments.remove(commitment);
			} else {
				planStep.handle(this, commitment);
				if (commitment.isFailed()) {
					if (!commitment.rollback()) {
						commitments.remove(commitment);
					}
				}
			}
		}
		
		// Signal the end of the iteration...
        this.endOfIteration();
        event = null;
        option = null;
	}
	
	public QueryResult queryAll(IFormula formula) {
		return queryEngine.queryAll(formula);
	}

	@Override
	/**
	 * Implement response to lack of an action
	 */
	protected void noSuchAction(Predicate activity) {
		String a = CoreUtilities.presenter.toString(activity);
		System.out.println("Unknown Action Call: " + a);
		Predicate predicate = (Predicate) CoreUtilities.factory.createFormula("unknownActivity(" + a + ")[source(percept)]");
		this.addEvent(new ASEvent(ASEvent.ADDITION, predicate));
	}
	
	public ASEvent getEvent() {
		return event;
	}
	
	public List<PlanRule> getOptions() {
		return options;
	}
	
	public PlanRule getSelected() {
		return option;
	}
	
	public List<Predicate> getBeliefs() {
		return beliefs.getBeliefs();
	}
	
	public List<ASEvent> getEventQueue() {
		return eventQueue;
	}
	
	public List<Commitment> getCommitments() {
		return commitments;
	}
	
	public List<PlanRule> getRules() {
		List<PlanRule> list = new LinkedList<PlanRule>();
		for(List<PlanRule> rules: ruleEventMap.values()) {
			list.addAll(rules);
		}
		return list;
	}
	
	public Collection<Action> getActions() {
		return actionMap.values();
	}

	public Collection<Sensor> getSensors() {
		return sensorMap.values();
	}

	@Override
	public String getType() {
		return "AF-AgentSpeak";
	}

	@Override
	public void initialise(String data) {
		if (data.startsWith("+")) {
			Predicate event = (Predicate) Utilities.factory.createFormula(data.substring(1));
			addEvent(new ASEvent(ASEvent.ADDITION, event));
		} else if (data.startsWith("-")) {
			Predicate event = (Predicate) Utilities.factory.createFormula(data.substring(1));
			addEvent(new ASEvent(ASEvent.REMOVAL, event));
		} else {
			try {
				addBelief((Predicate) Utilities.factory.createFormula(data+"[source(self)]"));
			} catch (LogicException le) {
				System.out.println("[" + getName() + "] Initialisation Error: " + le.getMessage());
			}
		}
	}

	@Override
	public void update(Observable arg0, Object arg1) {
		// TODO Auto-generated method stub
	}
	
	public void addBelief(Predicate belief) {
		addBelief(belief, null);
	}

	public void addBelief(Predicate belief, Object source) {
		if (beliefs.addBelief(belief)) {
			addEvent(new ASEvent(ASEvent.ADDITION, belief, source));
		}
	}
		
	public void removeBelief(Predicate belief) {
		removeBelief(belief, null);
	}
	
	public void removeBelief(Predicate belief, Object source) {
		if (beliefs.removeBelief(belief)) {
			addEvent(new ASEvent(ASEvent.REMOVAL, belief, source));
		}
	}
	
	public void addEvent(ASEvent event) {
		eventQueue.add(event);
	}
	
	protected void addPlanRule(PlanRule rule) {
		List<PlanRule> list = ruleEventMap.get(Utilities.createSignature(rule.getTriggeringEvent().content()));
		if (list == null) {
			list = new LinkedList<PlanRule>();
			ruleEventMap.put(Utilities.createSignature(rule.getTriggeringEvent().content()), list);
		}
		list.add(rule);
		eventSelector.registerPlanRule(rule);
	}
	
	public IQueryEngine getQueryEngine(){
		return queryEngine;
	}
}
