package com.agentfactory.logic.agent;

import com.agentfactory.logic.MentalAgent;
import com.agentfactory.logic.lang.FOS;
import com.agentfactory.platform.core.IPlatformService;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
/**
 * Perceptors are used by the agent to maintain an up to date model of its environment
 * in the form of beliefs.  All perceptors must subclass this class.
 *
 * @author Terry Lowen
 */
public abstract class Perceptor {
    protected MentalAgent agent;
    private List<FOS> perceptions;
    private FOS identifier;
    private Map configuration;
    private List ontologies;
    
    public Perceptor() {
        perceptions = new ArrayList();
    }
    
    public void setAgent(MentalAgent agent) {
        this.agent = agent;
    }
    
    public String getAgentName() {
        return agent.getName();
    }
    
    public void setConfiguration(Map configuration) {
        this.configuration = configuration;
    }
    
    public Map getConfiguration() {
        return configuration;
    }
    
    public void setIdentifier(FOS identifier) {
        this.identifier = identifier;
    }
    
    public FOS getIdentifier() {
        return identifier;
    }
    
    public void setOntologies(List ontologies) {
        this.ontologies = ontologies;
    }
    
    public Belief createBelief(String functor) {
        FOS term = null;
        Ontology ontology = null;
        Iterator it = ontologies.iterator();
        while (it.hasNext()) {
            ontology = agent.getOntologyManager().getOntology((String) it.next());
            term = ontology.getPredicate(functor);
            if (term != null) {
                return new Belief(term);
            }
        }
        return null;
    }
    
    public Belief createBelief(String functor, String ontology) {
        if (ontologies.contains(ontology)) {
            Ontology ont = agent.getOntologyManager().getOntology(ontology);
            FOS term = ont.getPredicate(functor);
            if (term != null) {
                return new Belief(term);
            }
        }
        
        return null;
    }
    
    /***************************************************************************
     * BELIEF UPDATE METHODS
     ***************************************************************************/
    /**
     * causes the agent to adopt the given belief
     *
     * @param bel the belief to be adopted
     */
    public void adoptBelief(String bel) {
        perceptions.add(new FOS(bel));
    }
    
    public void adoptBelief(Belief belief) {
        perceptions.add(belief.toBelief());
    }
    
    public IPlatformService getService(String name) {
        return agent.getPlatformService(name);
    }
    
    public List<IPlatformService> getServicesByClass(Class targetClass) {
        return agent.getPlatformServicesByClass(targetClass);
    }
    
    public Module getModuleByClass(String className) {
        return agent.getModuleByClass(className);
    }
    
    public Module getModuleByName(String name) {
        return agent.getModuleByName(name);
    }
    
    public void fire() {
        long startTime = System.currentTimeMillis();
        perceptions.clear();
        perceive();
        long endTime = System.currentTimeMillis();
        //updatePerformance(endTime-startTime);
    }
            
    /***************************************************************************
     * AGENT PERCEPTION METHOD
     ***************************************************************************/
    /**
     * executes each itertation of agent. should extract some knowledge from environment and
     * generate appropriate beliefs
     *
     */
    public abstract void perceive();
    
    /**
     * Destroy method useful for making sure that resuources used by agents
     * are dealocated before the agent is killed.
     *
     * Perceptors should implement this method if they have resources
     * which should be explictly desposed of when the agent is being killed
     * (e.g. native resources). Otherwise the default implementation
     * will suffice.
     */
    public void destroy() {
    }

    @Override
    public String toString() {
        return getClass().getName();
    }
    
    public List<FOS> getPerceptions() {
        return perceptions;
    }
}

