package com.agentfactory.logic.lang;

import com.agentfactory.logic.reasoner.SubstitutionSet;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * This class represents First Order Structures (FOS)
 *
 * @author Terry Lowen, Rem Collier
 * @version July 2002
 */
public class FOS implements Cloneable,Serializable,Iterable<FOS> {
    private String functor;
    private List<FOS> arguments;
    
    /**
     * creates a new empty FOS
     *
     */
    public FOS(String functor, List<FOS> arguments) {
        this.functor = functor;
        this.arguments = arguments;
        Logger.getLogger(functor).setLevel(Level.SEVERE);
    }
    
    public boolean isNot() {
        return functor.charAt(0) == '!' || functor.charAt(0) == '~';
    }
    
    /**
     * A nasty but necessarly constructor
     *
     * TODO - RR - Solve bug which allows for FOSs to be created badly out of
     * invalid strings.
     */
    public FOS(String inString) {
        // Ask the logic builder to build a FOS
        try {
            FOS temp = (FOS) LogicFactory.create(inString);
            functor = temp.functor;
            arguments = temp.arguments;
            Logger.getLogger(functor).setLevel(Level.SEVERE);
        } catch (MalformedLogicException mle) {
            functor = mle.getMessage();
        }
    }
    
    //---------------accessing
    
    /**
     * gets arguments
     *
     * @return FOS arguments
     */
    public List<FOS> getArguments(){
        return arguments;
    }

    public FOS[] arguments() {
        return (FOS[]) arguments.toArray();
        
    }
    
    /**
     * gets the number of arguments
     *
     * @return number of arguments
     */
    public int numArguments() {
        if (arguments == null) return 0;
        return arguments.size();
    }
    
    
    /**
     * gets argument at specified position
     *
     * @param i position of argument
     * @return argument at specified position
     */
    public FOS argAt(int i){
        return (FOS)arguments.get(i);
    }
    
    /**
     * get functor name
     *
     * @return name of functor
     */
    public String getFunctor(){
        if (functor.contains("\"")) {
            return functor.substring(1,functor.length()-1);
        }
        return functor;
    }
    
    
    /**
     * get copy of this
     *
     * @return copy of this
     */
    public FOS copy() {
        return new FOS(getFunctor(), arguments);
    }
    
    //-----------comparing
    /**
     * checks to see if this is a rule
     *
     * @return false
     */
    public boolean isRule() {
        return false;
    }
    
    
    /**
     * checks to see if this is Composite (has arguments)
     *
     * @return true if composite
     */
    public boolean isComposite(){
        return arguments.size() > 0;
    }
    
    
    /**
     * checks to see if this is a constant
     *
     * @return true if constant
     */
    public boolean isConstant(){
        return (arguments.size() == 0) && getFunctor().charAt(0) != '?';
    }
    
    
    /**
     * checks to see if this is a variable (starts with a '?')
     *
     * @return true if variable
     */
    public boolean isVariable() {
        Logger.getLogger(functor).log(Level.INFO, "isVariable");
        return (numArguments() == 0) && getFunctor().charAt(0) == '?';
    }
    
    /**
     * checks to see if this is a false ("false")
     *
     * @return true if false
     */
    public boolean isFalse(){
        if (isConstant()) {
            return (functor.equalsIgnoreCase("false"));
        } else if (functor.equals("BELIEF")) {
            return arguments.get(0).isFalse();
        }
        
        return false;
    }
    
    
    /**
     * checks to see if this is a true ("true")
     *
     * @return true if true
     */
    public boolean isTrue(){
        if (this.isConstant()) {
            return functor.equalsIgnoreCase("true");
        } else if (functor.equals("BELIEF")) {
            return arguments.get(0).isTrue();
        }
        
        return false;
    }
    
    
    //--comparing two FOS------
    
    /**
     * compares this with another FOS to see if they are exactly the same
     *
     * @param f the formula to be compared
     * @return true if exactly the same
     */
    public boolean isSameAs(FOS f) {
        if(!getFunctor().equals(f.getFunctor())) {
            return false;
        }
        
        if(arguments.size() != f.arguments.size()) {
            return false;
        }
        
        int i = 0;
        boolean matches = true;
        while (matches && (i < arguments.size())) {
            matches = arguments.get(i).isSameAs(f.arguments.get(i));
            i++;
        }
        
        return matches;
    }
    
    
    /**
     * Check if this FOS matches another FOS
     *
     * @param f formula to be matched
     * @return true if they match
     */
    public boolean matches(FOS f) {
        if ((f.isVariable()) || isVariable()) {
            return true;
        }
        
        if (!getFunctor().equals(f.getFunctor())) {
            Logger.getLogger(functor).log(Level.INFO, "Failure to match: " + functor + ":" + f.functor);
            return false;
        }
        
        if (arguments.size() != f.arguments.size()) {
            Logger.getLogger(functor).log(Level.INFO, "Different argument size: " + functor + ":" + f.functor);
            return false;
        }
        
        int i = 0;
        boolean matches = true;
        while (matches && (i < arguments.size())) {
            matches = arguments.get(i).matches(f.arguments.get(i));
            i++;
        }
        
        return matches;
    }
    
    
    /**
     * checks to see if specified formula is distinct to this
     *
     * @param f formula to check
     * @return true if distinct constants
     */
    public boolean isDistinctConstant(FOS f) {
        Logger.getLogger(functor).log(Level.INFO, "isDistinctConstant");
        return isConstant() && f.isConstant() && !getFunctor().equals(f.getFunctor());
    }
    
    
    /**
     * checks to see if specified formula has a distinct function to this
     *
     * @param f formula to check
     * @return true if distinct functions
     */
    public boolean isDistinctFunction(FOS f){
        Logger.getLogger(functor).log(Level.INFO, "isDistinctFunction");
        return isComposite() && f.isComposite() && !getFunctor().equals(f.getFunctor());
    }
    
    /**
     * checks to see if specified formula is composite while this is a constant or vice versa
     *
     * @param f formula to check
     * @return true if different fomulae
     */
    public boolean isDifferentTo(FOS f) {
        Logger.getLogger(functor).log(Level.INFO, "isDifferentTo");
        return (isConstant() && f.isComposite()) || (isComposite() && f.isConstant());
    }
    
    
    /**
     * formulae are compatible if they have the same functor and same number of arguments
     *
     * @param f formula to check
     * @return true if compatible functions
     */
    public boolean isCompatibleFunction(FOS f) {
        Logger.getLogger(functor).log(Level.INFO, "IsCompatibleFunction");
        return isComposite() && f.isComposite() && getFunctor().equals(f.getFunctor());
    }
    
    
    /**
     * formulae are not functional if one is a variable contained by the other
     *
     * @param f formula to check
     * @return true if not functional
     */
    public boolean isNotFunctional(FOS f) {
        Logger.getLogger(functor).log(Level.INFO, "isNotFunctional");
        return (isVariable() && f.containsVariable(this)) || (f.isVariable() && containsVariable(f));
    }
    
    
    
    //MAY NOT WORK PROPERLY
    /**
     * checks if specified variable is contained within this
     *
     * @param f variable to be checked
     * @return true if this contains variable
     */
    private boolean containsVariable(FOS variable){
        boolean found = false;
        
        int i=0;
        FOS arg = null;
        while (!found && i<arguments.size()) {
            arg = arguments.get(i++);
            found = arg.isSameAs(variable) || (arg.containsVariable(variable));
        }
        
        return found;
    }
    
    
    /**
     * applies a SubstitutionSet to this
     *
     * @param sub the SubstitutionSet to be used
     * @return a formula that has had the SubstitutionSet applied to it
     */
    public FOS apply(SubstitutionSet sub) {
        if( isVariable() ) {
            return sub.findConstant(this);
        }
        
        List<FOS> args = new ArrayList<FOS>();
        for(FOS fos:arguments) {
            //Logger.getLogger(functor).log(Level.INFO, "Appling: " + fos);
            if(fos.isConstant()) {
                args.add(fos);
            } else {
                args.add(fos.apply(sub));
            }
        }
        
        return new FOS(getFunctor(), args);
    }
    
    /**
     * gets a String representation of this
     *
     * @return this as a String
     */
    public String toString(){
        StringBuffer aString = new StringBuffer();
        aString.append(getFunctor());
        
        if (!arguments.isEmpty()) {
            aString.append("(");
            boolean first = true;
            for (FOS arg: arguments) {
                if(first) {
                    first = false;
                } else {
                    aString.append(",");
                }
                aString.append(arg.toString());
            }
            aString.append(")");
        }
        
        return aString.toString();
    }
    
    /**
     * Clone the current object
     */
    @Override
    public Object clone() {
        List<FOS> args = new ArrayList<FOS>();
        
        // if there are children call on each of these to copy
        for(FOS arg: arguments) {
            args.add((FOS) arg.clone());
        }
        
        // copy the functor
            new String(getFunctor());
        
        return new FOS(new String(getFunctor()), args);
    }
    
    @Override
    public boolean equals(Object object) {
        if (object instanceof FOS) {
            FOS fos = (FOS) object;
        
            if (!getFunctor().equals(fos.getFunctor())) {
                return false;
            }
        
            if (arguments.size() != fos.arguments.size()) {
                return false;
            }
        
            if (arguments.isEmpty()) {
                return true;
            }
        
            boolean validArguments = true;
            
            int i = 0;
            while(validArguments && (i < arguments.size())) {
                validArguments = fos.arguments.get(i).equals(arguments.get(i));
                i++;
            }
        
            return validArguments;
        }
        
        return false;
    }
    
    public List<FOS> getVariables() {
        List<FOS> list = new ArrayList<FOS>();
        if (isVariable()) {
            list.add(this);
        } else if (isComposite()) {
            for (FOS arg: arguments) {
                list.addAll(arg.getVariables());
            }
        }
        return list;
    }

    public Iterator<FOS> iterator() {
        return arguments.iterator();
    }
}


