/**
 * Title:       LogicFactory.java
 * Copyright:   Copyright (c) 1996-2004 The Agent Factory Working Group. All rights reserved.
 * Licence:     This file is free software; you can redistribute it and/or modify
 *              it under the terms of the GNU Lesser General Public License as published by
 *              the Free Software Foundation; either version 2.1, or (at your option)
 *              any later version.
 *
 *              This file is distributed in the hope that it will be useful,
 *              but WITHOUT ANY WARRANTY; without even the implied warranty of
 *              MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *              GNU Lesser General Public License for more details.
 *
 *              You should have received a copy of the GNU Lesser General Public License
 *              along with Agent Factory; see the file COPYING.  If not, write to
 *              the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *              Boston, MA 02111-1307, USA.
 */
package com.agentfactory.logic.lang;

import com.agentfactory.logic.lang.TerAND;
import com.agentfactory.logic.lang.MalformedLogicException;
import java.util.ArrayList;
import java.util.List;

/**
 * This is a factory class that contains methods that support the creation
 * of logical items.
 *
 * @author  Rem Collier
 */
public class LogicFactory {
    private static final char COMMA                                      = ',';
    private static final char QUOTE                                      = '"';
    private static final char OPEN_BRACKET                               = '(';
    private static final char CLOSE_BRACKET                              = ')';
    
    /**
     * Creates an object that represents the Agent Factory code contained
     * within the line parameter.
     *
     * @param line a String representation of the code.
     * @return an object representation of the code.
     */
    public static Object create(String line) throws MalformedLogicException {
        Object object = buildImplication(line);
        if (object == null) object = buildAnd(line);
        if (object == null) object = buildFOS(line);
        
        return object;
    }
    
    private static int getIndex(String string, char target) {
        return getIndex(string, target, 0);
    }
    
    private static int getIndex(String string, char target, int offset) {
        int bracketCount = 0;
        int index = offset;
        char character = target;
        boolean loop = true;
        boolean success = false;
        
        if (index >= string.length()) {
            loop = false;
        }
        
        while (loop) {
            character = string.charAt(index);
            if ((character == target) && (bracketCount == 0)) {
                loop = false;
                success = true;
            } else {
                if (character == '(') bracketCount++;
                if (character == ')') bracketCount--;
                index++;
                if (index == string.length()) {
                    loop = false;
                }
            }
        }
        
        if (!success) return -1;
        
        return index;
    }
    
    private static TerImplication buildImplication(String string) throws MalformedLogicException {
        int index = string.indexOf("=>");
        if (index == -1) {
            return null;
        }

        Object left = create(string.substring(0, index).trim());
        if (left instanceof FOS) {
            TerAND and = new TerAND();
            and.add((FOS)left);
            left = and;
        }
        
        FOS right = null;
        try {
            right = (FOS) create(string.substring(index+2).trim());
        } catch (ClassCastException cce) {
            throw new MalformedLogicException("Right hand side of " + string + " must be an FOS");
        }
        
        return new TerImplication((TerAND) left, right);
    }
    
    
    /**
     * Given a string which contains an arguement which apparently starts at
     * offset, find the index of the last character of this arguement.
     * Return -1 if there was no valid arguement could be found in this string
     */
    private static int getArgument(String string, int offset) throws MalformedLogicException {
        int bracketCount = 0;
        int index = offset;
        char character = COMMA;
        boolean loop = true;
        boolean success = false;
        boolean inQuote = false;
        
        if (index >= string.length()) {
            loop = false;
        }
        
        while (loop) {
            character = string.charAt(index);
            if (character == QUOTE) {
                if(!inQuote){
                    inQuote = !inQuote;
                    index++;
                }
                else{
                    // we are in quotes already, check if that was escaped
                    // i.e. it is a part of the sentence rather than denoting the
                    // end of the sentence
                    char previousChar = string.charAt(index - 1);
                    if(previousChar == '\\'){
                        // we are still in quotes, proceed
                        index++;
                    }
                    else{
                        // that must have been the end of the string
                        inQuote = !inQuote;
                        index++;
                        
                        if(index == (string.length())){
                            // there is nothing more to be found
                            // counter intuitive, but say that success is false
                            loop = false;
                            success = false;
                        }
                    }
                }
            } else {
                if (!inQuote) {
                    if ((character == COMMA) && (bracketCount == 0)) {
                        loop = false;
                        success = true;
                    } else {
                        if (character == OPEN_BRACKET) bracketCount++;
                        if (character == CLOSE_BRACKET) bracketCount--;
                        index++;
                        if (index == string.length()) {
                            loop = false;
                        }
                    }
                } else {
                    index++;
                    if (index == string.length()) {
                        throw new MalformedLogicException("Unmatching quotes in: '" + string + "'");
                    }
                }
            }
        }
        
        if (!success) return -1;
        
        return index;
    }
    
    private static FOS buildFOS(String string) throws MalformedLogicException {
        String functor = null;
        List<FOS> arguments = new ArrayList<FOS>();
        
        int startIndex = string.indexOf('(');
        
        if (startIndex == -1) {
            functor = string.trim();
        }
        else if(string.charAt(0) == '\"'){
            functor = (string.trim());
        }
        else {
            functor = (string.substring(0, startIndex).trim());
            
            // Start by getting the contents of the FOS
            int endIndex = string.lastIndexOf(')');
            if (endIndex == -1) {
                throw new MalformedLogicException("Failed to build: " + string);
            }
            
            string = string.substring(startIndex+1, endIndex).trim();
            startIndex = 0;
            
            // while it appears that there is more stuff to be found
            while ((endIndex = getArgument(string, startIndex)) > -1) {
                // if we are getting to this point, we are assuming that this
                // thing that we are adding is definately not the last thing
                // to be added
                arguments.add(buildFOS(string.substring(startIndex, endIndex).trim()));
                startIndex = endIndex + 1;
            }
            
            arguments.add(buildFOS(string.substring(startIndex).trim()));
        }
        
        return new FOS(functor, arguments);
    }
    
    private static TerAND buildAnd(String string) throws MalformedLogicException {
        TerAND and = new TerAND();
        
        int startIndex = 0;
        int endIndex = getIndex(string, '&');
        while (endIndex > -1) {
            and.add(buildFOS(string.substring(startIndex, endIndex).trim()));
            startIndex = endIndex + 1;
            endIndex = getIndex(string, '&', startIndex);
        }
        
        // Check if startIndex is still 0 (i.e. no occurrences of '&').
        if (startIndex == 0) {
            return null;
        }
        
        // Add last term
        and.add(buildFOS(string.substring(startIndex).trim()));
        
        return and;
    }
}


