package com.agentfactory.logic.update.interpreter;

import java.util.HashMap;
import java.util.Map;

import com.agentfactory.logic.update.lang.IFormula;
import com.agentfactory.logic.update.lang.Predicate;
import com.agentfactory.logic.update.reasoner.QueryResult;
import com.agentfactory.platform.core.AgentID;
import com.agentfactory.platform.core.IMessage;
import com.agentfactory.platform.core.IPlatformService;
import com.agentfactory.platform.impl.AbstractAgent;
import com.agentfactory.platform.impl.StringMessage;

public abstract class AbstractLanguageAgent extends AbstractAgent {
	protected Map<String, Action> actionMap;
	protected Map<String, Sensor> sensorMap;
	protected Map<String, Module> moduleMap;

	public AbstractLanguageAgent(String name) {
		super(name);
		
		actionMap = new HashMap<String, Action>();
		sensorMap = new HashMap<String, Sensor>();
		moduleMap = new HashMap<String, Module>();
	}
	
	/**
	 * Default Sensing Algorithm:
	 * 
	 * 1. agent level sensing
	 * 2. module level sensing
	 * 
	 */
	protected void senseEnvironment() { 
		// Part A: Sense the environment...
		for (Sensor sensor: sensorMap.values()) {
			sensor.perceive();
		}
        for (Module module: moduleMap.values()) {
    		for (Sensor sensor: module.getSensors()) {
    			sensor.perceive();
    		}
        }
	}

	protected void noSuchAction(Predicate activity) {
	}
	
	/**
	 * Invoke the specified action:
	 * 
	 * 1. attempt an agent level action
	 * 2. attempt a module level action
	 * 3. if no action, then call the customisable noSuchAction() method
	 * 
	 * @param activity
	 * @return
	 */
	public boolean performActivity(Predicate activity) {
		Action action = actionMap.get(CoreUtilities.createSignature(activity));
		if (action == null) {
			action = getModuleLevelAction(activity);
			if (action == null) {
				noSuchAction(activity);
				return false;
			}
		}
		return action.execute(activity);
	}

	/**
	 * Retrieves a module level action or returns null if none
	 * exists.
	 * 
	 * @param activity
	 * @return
	 */
	private Action getModuleLevelAction(Predicate activity) {
		String sig = CoreUtilities.createSignature(activity);
		int index = sig.indexOf(".");
		if (index == -1) return null;
		
		String id = sig.substring(0, index) +":0";
		Module module = moduleMap.get(id);
		if (module == null) {
			System.out.println("No Module: " + id);
			return null;
		}
		return module.getAction(sig.substring(index+1));
	}

	public void addModule(String id, Module module) {
		addModule((Predicate) CoreUtilities.factory.createFormula(id), module);
	}
	
	public void addModule(Predicate id, Module module) {
		module.setId(id);
		module.setAgent(this);
		moduleMap.put(CoreUtilities.createSignature(id), module);
	}
	
	protected void addAction(String name, Action action) {
		action.setId((Predicate) CoreUtilities.factory.createFormula(name));
		action.setAgent(this);
		actionMap.put(CoreUtilities.createSignature(action.getId()), action);
	}
	
	protected void addSensor(String name, Sensor sensor) {
		sensor.setId((Predicate) CoreUtilities.factory.createFormula(name));
		sensor.setAgent(this);
		sensorMap.put(CoreUtilities.createSignature(sensor.getId()), sensor);
	}
	
	public Module getModule(Predicate module) {
		return moduleMap.get(CoreUtilities.createSignature(module));
	}
	
	public abstract void addBelief(Predicate belief); 
	public abstract void removeBelief(Predicate belief);

	protected void installBasicAPI() {
		addAction(".abolish(?belief)", new Action() {
			@Override
			public boolean execute(Predicate activity) {
				String bel = CoreUtilities.presenter.toString(activity.termAt(0));
				removeBelief(bel);
				return true;
			}
		});
		addAction(".send(?perf, ?rec, ?content)", new Action() {
			@Override
			public boolean execute(Predicate activity) {
				String performative = CoreUtilities.presenter.toString(activity.termAt(0));
				AgentID receiver = CoreUtilities.toAgentID(activity.termAt(1));
				String content = CoreUtilities.presenter.toString(activity.termAt(2));
	
		        StringMessage message = StringMessage.newInstance();
		        message.setPerformative(performative.toUpperCase());
		        message.setSender(agent.getAgentID());
		        message.getReceivers().add(receiver);
		        message.setContent(content.toString());
		        message.setLanguage("AFAPL");
		        message.setOntology("none");
		        message.setProtocol("none");
		        
		        return agent.send(message);
			}
		});
		addAction(".print(?content)", new Action() {
			@Override
			public boolean execute(Predicate activity) {
				String content = CoreUtilities.presenter.toString(activity.termAt(0));
				System.out.print(content);
				return true;
			}
		});
		addAction(".println(?content)", new Action() {
			@Override
			public boolean execute(Predicate activity) {
				String content = CoreUtilities.presenter.toString(activity.termAt(0));
				System.out.println(content);
				return true;
			}
		});
		addAction(".fail", new Action() {
			@Override
			public boolean execute(Predicate activity) {
				return false;
			}
		});
		addAction(".nil", new Action() {
			@Override
			public boolean execute(Predicate activity) {
				return true;
			}
		});
		addAction(".bind(?id)", new Action() {
			@Override
			public boolean execute(Predicate activity) {
				String id = CoreUtilities.presenter.toString(activity.termAt(0));
				if (agent.bindToPlatformService(id)) {
					this.addBelief("boundToService(" + id + ")");
					return true;
				}
				addBelief("bindToServiceFailed(" + id + ")");
				return false;
			}
		});
		addAction(".createModule(?id, ?design)", new Action() {
			@SuppressWarnings("unchecked")
			@Override
			public boolean execute(Predicate activity) {
				Predicate id = CoreUtilities.factory.convertToPredicate(activity.termAt(0));
				String design = CoreUtilities.presenter.toString(activity.termAt(1));
				try {
					Class cl = Class.forName(design);
					Module module = (Module) cl.newInstance();
					addModule(id, module);
				} catch (InstantiationException e) {
					return false;
				} catch (IllegalAccessException e) {
					return false;
				} catch (ClassNotFoundException e) {
					return false;
				}
				return true;
			}
		});
		addSensor(".services", new Sensor() {
			@Override
			public void perceive() {
				for(IPlatformService p :agent.getBoundPlatformServices()){
					addBelief("service("+p.getName()+")");
				};			
			}
		});
		addSensor(".inbox", new Sensor() {
			@Override
			public void perceive() {
		        for (IMessage message: agent.getInbox()) {
		        	addBelief("message(" + message.getPerformative() + "," +
		        				  message.getSender().toFOSString() + "," +
		        				  message.getContent() + ")");
		        }
			}
		});
		addSensor(".self", new Sensor() {
			@Override
			public void perceive() {
		        addBelief("name(" + agent.getName() + ")");
		        addBelief(agent.getAgentID().toFOSString());
			}
		});
	}
	
	public abstract QueryResult queryAll(IFormula formula);
	
	public Map<String, Action> getActionMap() {
		return actionMap;
	}
	
	public Map<String, Sensor> getSensorMap() {
		return sensorMap;
	}
}
