package com.agentfactory.platform;

import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.agentfactory.platform.core.AgentCreationException;
import com.agentfactory.platform.core.DuplicateAgentNameException;
import com.agentfactory.platform.core.IAgent;
import com.agentfactory.platform.core.IArchitectureFactory;
import com.agentfactory.platform.core.NoSuchArchitectureException;
import com.agentfactory.platform.impl.DefaultAgentPlatform;
import com.agentfactory.platform.util.Scanner;

/**
 *
 * @author remcollier
 */
public class InitialisationManager {
    private static final String LOGNAME = InitialisationManager.class.getName();
    
    static {
        Logger.getLogger(LOGNAME).setLevel(Level.WARNING);
    }
    
    DefaultAgentPlatform platform;
    
    public InitialisationManager() {
        platform = new DefaultAgentPlatform();
    }
    
    public void configurePlatform(Scanner scanner) throws IOException {
        List<String> statement = null;
        while (scanner.hasNext()) {
            statement = scanner.nextStatement();
            if (!statement.isEmpty()) {
                String command = statement.get(0);
                if (command.equals("PLATFORM_NAME")) {
                    platform.setName(statement.get(1));
                } else if (command.equals("PLATFORM_DOMAIN")) {
                    platform.setDomain(statement.get(1));
                } else if (command.equals(("SERVICE"))) {
                    throw new UnsupportedOperationException("No Support for Services");
                } else if (command.equals(("SECURITY_POLICY"))) {
                    throw new UnsupportedOperationException("No Support for Services");
                } else if (command.equals(("ARCHITECTURE"))) {
                    try {
                        Class architecture = Class.forName(statement.get(1));
                        IArchitectureFactory factory = (IArchitectureFactory) architecture.newInstance();
                        Properties configuration = new Properties();
                        int i = 2;
                        while (i < statement.size()) {
                            String[] arg = statement.get(i).split("=");
                            configuration.setProperty( arg[0], arg[1] );
                        }
                        factory.configure(configuration);
                        platform.getArchitectureService().registerArchitectureFactory(factory);
                    } catch (ClassNotFoundException ex) {
                        Logger.getLogger(InitialisationManager.class.getName()).log(Level.SEVERE, null, ex);
                    } catch (InstantiationException ex) {
                        Logger.getLogger(InitialisationManager.class.getName()).log(Level.SEVERE, null, ex);
                    } catch (IllegalAccessException ex) {
                        Logger.getLogger(InitialisationManager.class.getName()).log(Level.SEVERE, null, ex);
                    }
                } else {
                    throw new UnsupportedOperationException("Invalid Statment: " + statement);
                }
            }
        }
    }
    
    public void initialiseCommunity(Scanner scanner) throws IOException {
        Map<String, IAgent> agents = new HashMap<String, IAgent>();
        
        List<String> statement = null;
        while (scanner.hasNext()) {
            statement = scanner.nextStatement();
            if (!statement.isEmpty()) {
                String command = statement.get(0);
                if (command.equals("SYSTEMAGENT")) {
                    try {
                        IAgent agent = platform.getArchitectureService().createAgent(statement.get(1), statement.get(2));
                        platform.getAgentContainer().add(agent);
                        platform.getScheduler().register(agent);
                        agents.put(agent.getName(), agent);
                    } catch (DuplicateAgentNameException ex) {
                        Logger.getLogger(InitialisationManager.class.getName()).log(Level.SEVERE, null, ex);
                    } catch (NoSuchArchitectureException ex) {
                        Logger.getLogger(InitialisationManager.class.getName()).log(Level.SEVERE, null, ex);
                    } catch (AgentCreationException ex) {
						Logger.getLogger(InitialisationManager.class.getName()).log(Level.SEVERE, null, ex);
					}
                } else if (command.equals("CREATE_AGENT")) {
                    try {
                        IAgent agent = platform.getArchitectureService().createAgent(statement.get(1), statement.get(2));
                        platform.getAgentContainer().add(agent);
                        platform.getScheduler().register(agent);
                        agents.put(agent.getName(), agent);
                    } catch (DuplicateAgentNameException ex) {
                        Logger.getLogger(InitialisationManager.class.getName()).log(Level.SEVERE, null, ex);
                    } catch (NoSuchArchitectureException ex) {
                        Logger.getLogger(InitialisationManager.class.getName()).log(Level.SEVERE, null, ex);
                    } catch (AgentCreationException ex) {
						Logger.getLogger(InitialisationManager.class.getName()).log(Level.SEVERE, null, ex);
                    }
                } else if (command.equals("INITIALISE")) {
                    agents.get(statement.get(1)).initialise(remainder(statement, 2));
                } else if (command.equals("START_AGENT")) {
                    platform.getScheduler().resume(statement.get(1));
                } else {
                    throw new UnsupportedOperationException("Invalid Statment: " + statement);
                }
            }
        }
    }
    
    private String remainder(List<String> statement, int index) {
        StringBuffer buf = new StringBuffer();
        
        boolean first = true;
        for (int i = index; i < statement.size(); i++) {
            if (first) first = false;
            else buf.append(" ");
            buf.append(statement.get(i));
        }
        
        return buf.toString();
    }
    
    public static void main(String[] args) {
        String script = null;
        String configFile = null;

        for (int i = 0; i < args.length; i++) {
            String nextString = args[i];
            if (nextString.equals("-script")) {
                script = args[++i];
            } else if (nextString.equals("-config")) {
                // start the platform
                configFile = args[++i];
            } else {
                throw new UnsupportedOperationException("Unrecognised input: " + args[i]);
            }
        }

        InitialisationManager manager = new InitialisationManager();
        try {
            manager.configurePlatform(new Scanner(InitialisationManager.class.getResourceAsStream("/" + configFile)));
            manager.initialiseCommunity(new Scanner(InitialisationManager.class.getResourceAsStream("/" + script)));
        } catch (IOException ex) {
            Logger.getLogger(InitialisationManager.class.getName()).log(Level.SEVERE, null, ex);
        }
    }
}
