package com.agentfactory.platform.impl;

import com.agentfactory.platform.core.AgentID;
import com.agentfactory.platform.core.IEnvelope;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import org.w3c.dom.CharacterData;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

/**
 * This is a concrete implementation of the Envelope class that parses
 * some text that contains an XML version of the envelope.
 *
 * @author  Rem Collier
 */
public class XMLEnvelope implements IEnvelope {
    AgentID from;
    List<AgentID> to;
    List<AgentID> intendedReceiver;
    Calendar date;
    String payloadEncoding;
    long payloadLength;
    String aclRepresentation;
    
    public XMLEnvelope() {
        to = new LinkedList<AgentID>();
        intendedReceiver = new LinkedList<AgentID>();
    }

    /**
     * This factory method returns an ArrayList of XMLEnvelope objects - one
     * per envelope defined in the message.
     *
     * @param text the XML in string format
     * @return an ArrayList of XMLEnvelope objects!
     */
    public static List newInstance(String text) {
        List envelopes = new ArrayList();
        DocumentBuilder builder;
        try {
            builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
            Document doc = builder.parse(new InputSource(new StringReader(text)));
            NodeList nodes = doc.getElementsByTagName("params");
            
            XMLEnvelope envelope = null;
            for (int i=0; i<nodes.getLength(); i++) {
                envelope = new XMLEnvelope();
                NodeList envelopeNodes = nodes.item(i).getChildNodes();

                for (int j=0; j<envelopeNodes.getLength(); j++) {
                    Node node = envelopeNodes.item(j);
                    if (node.getNodeName().equals("to")) {
                        NodeList agentIdentifiers = node.getChildNodes();
                        for (int k=0; k<agentIdentifiers.getLength(); k++) {
                            if (agentIdentifiers.item(k).getNodeName().equals("agent-identifier")) {
                                envelope.to.add(parseAgentIdentifier(agentIdentifiers.item(k)));
                            }
                        }
                    } else if (node.getNodeName().equals("from")) {
                        NodeList agentIdentifiers = node.getChildNodes();
                        for (int k=0; k<agentIdentifiers.getLength(); k++) {
                            if (agentIdentifiers.item(k).getNodeName().equals("agent-identifier")) {
                                envelope.from = parseAgentIdentifier(agentIdentifiers.item(k));
                            }
                        }
                    } else if (node.getNodeName().equals("intended-receiver")) {
                        NodeList agentIdentifiers = node.getChildNodes();
                        for (int k=0; k<agentIdentifiers.getLength(); k++) {
                            if (agentIdentifiers.item(k).getNodeName().equals("agent-identifier")) {
                                envelope.intendedReceiver.add(parseAgentIdentifier(agentIdentifiers.item(k)));
                            }
                        }
                    } else if (node.getNodeName().equals("acl-representation")) {
                        //NodeList agentIdentifiers = node.getChildNodes();
                        envelope.aclRepresentation = ((CharacterData) node.getFirstChild()).getData();
                    } else if (node.getNodeName().equals("payload-length")) {
                        //NodeList agentIdentifiers = node.getChildNodes();
                        envelope.payloadLength = Integer.parseInt(((CharacterData) node.getFirstChild()).getData());
                    }
                }
                
                envelopes.add(envelope);
            }
        } catch (Exception ex) {
            Logger.getLogger(XMLEnvelope.class.getName()).log(Level.SEVERE, null, ex);
        }
        
        return envelopes;
    }

    private static AgentID parseAgentIdentifier(Node node) throws Exception {
        String name = null;
        List<String> addresses = new LinkedList<String>();
        List<String> resolvers = new LinkedList<String>();

        
        NodeList params = node.getChildNodes();
        for (int i=0; i<params.getLength(); i++) {
            Node param = params.item(i);
            if (param.getNodeName().equals("name")) {
                name = ((CharacterData) param.getFirstChild()).getData();
            }
            if (param.getNodeName().equals("addresses")) {
                NodeList urls = param.getChildNodes();
                for (int j=0; j<urls.getLength(); j++) {
                    if (urls.item(j).getNodeName().equals("url")) {
                        addresses.add(((CharacterData) urls.item(j).getFirstChild()).getData());
                    }
                }
            }
        }
        
        return new AgentID(name, addresses, resolvers);
    }

    /**
     * Create a String that containing the message envelope in a form that
     * is compliant with the FIPA XML Specification.
     *
     * @return an xml String
     */
     @Override
    public String toString() {
        StringBuffer buf = new StringBuffer();
        buf.append("<?xml version=\"1.0\"?>\n");
        buf.append("<envelope>\n");
        buf.append("\t<params index=\"1\">\n");
        buf.append("\t\t<to>\n");
        
        AgentID aid = null;
        for (int i = 0; i < to.size(); i++) {
            aid = (AgentID) to.get(i);
            aid.toXMLString(buf, "\t\t\t");
        }
        buf.append("\t\t</to>\n");
        buf.append("\t\t<from>\n");
        from.toXMLString(buf, "\t\t\t");
        buf.append("\t\t</from>\n");
        buf.append("\t\t<acl-representation>" + aclRepresentation + "</acl-representation>\n");
        buf.append("\t\t<payload-length>" + payloadLength + "</payload-length>\n");
        buf.append("\t\t<intended-receiver>\n");
        
        for (int i = 0; i < intendedReceiver.size(); i++) {
            aid = (AgentID) intendedReceiver.get(i);
            aid.toXMLString(buf, "\t\t\t");
        }
        buf.append("\t\t</intended-receiver>\n");
        buf.append("\t</params>\n");
        buf.append("</envelope>");
        return buf.toString();
    }
    
    public List<AgentID> getIntendedReceiver() {
        return intendedReceiver;
    }

    public void setIntendedReceiver( List<AgentID> recipients ) {
        this.intendedReceiver = recipients;
    }
    
    public List<AgentID> getTo() {
        return to;
    }
    
    public void setTo( List<AgentID> recipients ) {
        this.to = recipients;
    }

    public AgentID getFrom() {
        return from;
    }
    
    public void setFrom( AgentID sender ) {
        this.from = sender;
    }

    public String getAclRepresentation() {
        return aclRepresentation;
    }
    
    public void setAclRepresentation( String aclRepresentation ) {
        this.aclRepresentation = aclRepresentation;
    }

    public Calendar getDate() {
        return date;
    }
    
    public void setDate( Calendar date ) {
        this.date = date;
    }

    public String getPayloadEncoding() {
        return payloadEncoding;
    }
    
    public void setPayloadEncoding( String payloadEncoding ) {
        this.payloadEncoding = payloadEncoding;
    }

    public long getPayloadLength() {
        return payloadLength;
    }
    
    public void setPayloadLength(long payloadLength) {
        this.payloadLength = payloadLength;
    }
}
