package com.agentfactory.service.mts.udp;

import com.agentfactory.platform.core.IMessage;
import com.agentfactory.platform.core.TransportMessage;
import com.agentfactory.platform.impl.MessageTransportService;
import com.agentfactory.platform.impl.XMLEnvelope;
import java.util.Calendar;
import java.net.*;
import java.io.*;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Implementation of the FIPA HTTP Message Transport Service.
 *
 * @author  Rem Collier
 */
public class UDPMessageTransportService extends MessageTransportService {
    public static final String NAME = "fipa.mts.mtp.udp.std";
    
    private int port;
    private String hostname;
    private DatagramSocket socket;
    private InetAddress address;
    private UDPReceiver receiver;
    private String iface;
    
    @Override
    public void onStart() {
        port = Integer.parseInt(configuration.getProperty("port"));
        hostname = configuration.getProperty("host");
        
        try {
            address = InetAddress.getByName(hostname);
            if (address.isMulticastAddress()) {
                socket = new MulticastSocket();
                ((MulticastSocket)socket).joinGroup(address);
            }
        } catch (UnknownHostException uhe) {
            System.err.println("[UDPMessageTransportService] Unknown host: " + hostname);
        } catch (IOException ioe) {
            System.err.println("[UDPMessageTransportService] Error occurred setting up UDP connection for: " + hostname);
        }
            
        // Step 2: Set up a thread to handle incoming messages
        try {
            receiver = new UDPReceiver(this);
            receiver.start();
        } catch (IOException ioe) {
            System.err.println("[UDPMessageTransportService] Error occurred setting receiver UDP connection for: " + hostname);
        }
        
        Logger.getLogger(NAME).log(Level.WARNING, "UDP-MTS Listening on: " + hostname +":" + port);
    }
    
    /**
     * Checks if a given address is valid for this service.
     *
     * @param address the address to be checked
     * @return boolean -  true if the address is valid for the service. false otherwise.
     */
    public boolean isValidService(String address) {
        // TODO (Rem Collier): Need to make this stricter
        return address.startsWith("udp");
    }
    
    /**
     * This method is responsible for transmitting messages using the FIPA
     * HTTP Message Transport Service.
     *
     * @param address A String object
     * @param message a Message object
     */
    public boolean sendMessage(String address, IMessage message) {
        Logger.getLogger(NAME).log(Level.INFO, "Sending Message: " + message.toString());
        
        TransportMessage tmessage = createTransportMessage(message);
        
        byte[] buf = message.toString().getBytes();
        DatagramPacket packet = new DatagramPacket(buf, buf.length, this.address, port);
        try {
            socket.send(packet);
        } catch (IOException e) {
            return false;
        }
        
        return true;
    }
    
    public TransportMessage createTransportMessage(IMessage message) {
        byte[] byteArray = message.toByteArray();
        XMLEnvelope envelope = new XMLEnvelope();
        envelope.setTo(message.getReceivers());
        envelope.setFrom(message.getSender());
        envelope.setAclRepresentation(message.getAclRepresentation());
        envelope.setDate(Calendar.getInstance());
        envelope.setPayloadEncoding(message.getPayloadEncoding());
        envelope.setPayloadLength(byteArray.length);
        return new TransportMessage(envelope, message);
    }
    
    public String getAddress() {
        return "udp://" + hostname + ":" + port;
    }
    
    public String getServiceID() {
        return "udp-" + hostname + "-" + port;
    }

    public InetAddress getInetAddress() {
        return address;
    }
    
    public int getPort() {
        return port;
    }
    
    public String getInterface() {
        return iface;
    }
}
